/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * HRTF.cpp
 *
 *  Created on: 2011/03/01
 *      Author: HARK Development Team
 */

#include <stdexcept>
#include <HRTF.hpp>
#include <IIPDS.hpp>
#include <Utils.hpp>
#include <Logger.hpp>
using namespace std;
using namespace boost;

namespace NSHarkBinaural
{

namespace
{
std::vector<std::vector<harkio::TF_TYPE> >
SelectChannel(const std::vector<std::vector<harkio::TF_TYPE> >& tf, unsigned int l, unsigned int r);

void SetIIPDS(
    const std::vector<std::vector<harkio::TF_TYPE> >& tf,
    std::vector<double>& ipd, std::vector<double>& iid)
{
    if (tf.size() != 2)
    {
        throw length_error("tf.size != 2");
    }
    else if (tf[0].size() != tf[1].size())
    {
        throw length_error("tf[0].size != tf[1].size");
    }
    ipd.resize(tf[0].size());
    iid.resize(tf[0].size());
    double il, ir; // intencity l channel, r channel
    double pl, pr; // phase l channel, r channel
    harkio::TF_TYPE cl, cr;
    for (size_t i = 0; i < tf[0].size(); ++i)
    {
        cl = tf[0][i];
        cr = tf[1][i];
        il = Decibel(cl);
        ir = Decibel(cr);
        pl = arg(cl);
        pr = arg(cr);
        ipd[i] = pr - pl;
        iid[i] = ir - il;
    }
}

void SetIIPDS(
    const std::vector<std::vector<std::vector<std::vector<harkio::TF_TYPE> > > >& tf,
    IIPDS& iipds, unsigned int l, unsigned int r)
{
    iipds.ipd.resize(tf.size());
    iipds.iid.resize(tf.size());
    for (size_t dir = 0; dir < tf.size(); ++dir)
    {
        SetIIPDS(SelectChannel(tf[dir][0], l, r), iipds.ipd[dir].m, iipds.iid[dir].m);
    }
}

void SetIIPDS(
    const std::vector<std::vector<std::vector<harkio::TF_TYPE> > >& tf,
    IIPDS& iipds, unsigned int l, unsigned int r)
{
    iipds.ipd.resize(tf.size());
    iipds.iid.resize(tf.size());
    for (size_t dir = 0; dir < tf.size(); ++dir)
    {
        SetIIPDS(SelectChannel(tf[dir], l, r), iipds.ipd[dir].m, iipds.iid[dir].m);
    }
}

/** @brief Returns 2 channel transfer function created from the specified transfer function.
 * @param tf A transfer function which may have only one channel or more than 2 channels.
 * @param l The channel number of left. (must satisfy l < tf.size())
 * @param r The channel number of right. (must satisfy r < tf.size())
 * @return A 2 channel transfer function. (index 0 is tf[l], index 1 is tf[r])
 */
std::vector<std::vector<harkio::TF_TYPE> >
SelectChannel(const std::vector<std::vector<harkio::TF_TYPE> >& tf, unsigned int l, unsigned int r)
{
    if (tf.size() <= l || tf.size() <= r)
    {
        throw range_error("channel number is out of range");
    }
    std::vector<std::vector<harkio::TF_TYPE> > result(2);
    result[0] = tf[l];
    result[1] = tf[r];
    return result;
}

/**
 * 音源の配置が一層の円形になっているかどうか検査する.
 * @param pos 音源の配列 SRC_TYPE[Nd][Nr]
 * @return 配置が一層の円形ならtrue
 */
bool IsCircle(const std::vector<std::vector<harkio::SRC_TYPE> >& pos)
{
    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    if (pos.size() == 0) // 音源がない場合
    {
        LogFine(logger, "there is no source");
        return false;
    }
    else if (pos.size() == 1) // 音源の水平方向の種類が1つだけの場合
    {
        LogFine(logger, "there is one horizon kind");
        return pos[0].size() == 1 ? true : false;
    }

    // 一層になっているかどうか
    for (size_t i = 0; i < pos.size(); ++i)
    {
        if (pos[i].size() != 1)
        {
            LogFine(logger, "pos is not one layer");
            return false;
        }
    }

    const harkio::MIC_TYPE& v0 = pos[0][0];
    const harkio::MIC_TYPE& v1 = pos[1][0];

    // 半径を計算
    const double r0 = sqrt(v0.x * v0.x + v0.y * v0.y);

    // 半径の許容値（rに対する割合）
    const double rPermissible = 0.01;

    // すべての音源が半径rの円周上にあるか
    for (size_t i = 1; i < pos.size(); ++i)
    {
        const harkio::MIC_TYPE& v = pos[i][0];
        double r = sqrt(v.x * v.x + v.y * v.y);
        if (!(r0 * (1 - rPermissible) <= r && r <= r0 * (1 + rPermissible)))
        {
            LogFine(logger, "radiuses are not constant");
            return false;
        }
    }
    return true;
}

/**
 * 各音源の角度を計算する.
 * 各音源は一層の円形に並んでいると仮定する。
 * @param pos 音源の配列 SRC_TYPE[Nd][Nr]
 * @return 各音源の角度(degree)
 */
boost::shared_ptr<std::vector<double> >
CalcDirections(const std::vector<std::vector<harkio::SRC_TYPE> >& pos)
{
    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    shared_ptr<vector<double> > result(new vector<double>(pos.size()));
    vector<double>& dirs = *result;

    // 音源の角度を計算
    for (size_t i = 0; i < pos.size(); ++i)
    {
        const harkio::MIC_TYPE& v = pos[i][0];
        dirs[i] = 180 * atan2(v.y, v.x) / M_PI;
        LogFinest(logger, format("dirs[%d]=%f") % i % dirs[i]);
    }

    return result;
}

/**
 * 音源と音源の間の角度(degree)が一定かつ整数かどうか調べる.
 * @param dirs 各音源の角度(degree)
 * @return 間隔が一定かつ整数ならtrue
 */
bool IsFixedAndInteger(const std::vector<double>& dirs)
{
    Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    if (dirs.size() <= 1)
    {
        LogFine(logger, "the number of source directions is very little");
        return true;
    }

    int d10 = rint(dirs[1] - dirs[0]);
    int d0 = rint(dirs[0]);

    // d0とd1は境界をまたいでいるため、修正
    if (abs(d10) > 180)
    {
        d10 = rint(dirs[2] - dirs[1]);
        d0 = rint(dirs[1]) - d10;
    }

    for (size_t i = 1; i < dirs.size(); ++i)
    {
        int d = rint(dirs[i]);

        if (d0 + i * d10 != d)
        {
            LogFine(logger, format("expected:%d, actual:%d") % (d0 + i * d10) % d);
            return false;
        }
    }

    return true;
}

}

boost::shared_ptr<IIPDS> HRTFToIIPDS(harkio::HARK& harkFile, unsigned int channelSelectL, unsigned int channelSelectR)
{
    harkio::FMT* fmt = harkFile.getData()->getChunk<harkio::FMT>();

    if (fmt == 0)
    {
        throw invalid_argument("Invalid hark file. There is no FMT chunk");
    }

    // FMTチャンクのバージョン確認
    harkio::FMT::VERSION_CHECK versionCheck = fmt->checkVersion(harkio::FMT::V1_0);
    if (versionCheck == harkio::Chunk::MISMATCH)
    {
        throw invalid_argument("Invalid hark file. There is no version 1.0 FMT chunk");
    }

    shared_ptr<IIPDS> iipds(new IIPDS());
    iipds->fs = fmt->getNFS();
    iipds->nfft = fmt->getNFFT();
    iipds->nfreq = fmt->getNIR();

    // M2PGとGTFのどちらが格納されているか確認
    harkio::FMT::FORMAT_ID formatId = fmt->getFormatId();

    if (formatId == harkio::FMT::M2PG)
    {
        // M2PGチャンクからIIPDSを生成
        harkio::M2PG* m2pg = harkFile.getData()->getChunk<harkio::M2PG>();
        if (m2pg == 0)
        {
            throw invalid_argument("Invalid hark file. There is no M2PG chunk");
        }
        else if (m2pg->getTF() == 0)
        {
            throw invalid_argument("Invalid hark file. There is no TF");
        }
        //iipds->ndir = m2pg->getND();

        iipds->dirtick = 0;
        iipds->begindeg = 0;
        iipds->ndir = 0;

        if (IsCircle(*m2pg->getSrc()))
        {
            shared_ptr<vector<double> > dirsPtr = CalcDirections(*m2pg->getSrc());
            vector<double>& dirs = *dirsPtr;
            if (IsFixedAndInteger(dirs))
            {
                iipds->dirtick = dirs.size() >= 2 ? rint(dirs[1] - dirs[0]) : 0;
                iipds->begindeg = rint(dirs[0]);
                iipds->ndir = dirs.size();
            }
            else
            {
                throw invalid_argument("Invalid source position. (not fixed and integer tick)");
            }
        }
        else
        {
            throw invalid_argument("Invalid source position. (not a circle)");
        }

        SetIIPDS(*m2pg->getTF(), *iipds, channelSelectL, channelSelectR);
    }
    else if (formatId == harkio::FMT::GTF)
    {
        // GTFチャンクからIIPDSを生成
        harkio::GTF* gtf = harkFile.getData()->getChunk<harkio::GTF>();
        if (gtf == 0)
        {
            throw invalid_argument("Invalid hark file. There is no GTF chunk");
        }
        else if (gtf->getTF() == 0)
        {
            throw invalid_argument("Invalid hark file. There is no TF");
        }
        iipds->ndir = gtf->getNSrc();

        if (gtf->getNSrc() == 1)
        {
            iipds->dirtick = 0;
        }
        else if (gtf->getNSrc() >= 2)
        {
            iipds->dirtick = 180 / (gtf->getNSrc() - 1);
        }
        else
        {
            throw invalid_argument("The number of directions is invalid");
        }
        SetIIPDS(*gtf->getTF(), *iipds, channelSelectL, channelSelectR);
    }
    return iipds;
}

}
